// 9:02 AM 2/26/2023	- Changed the way multiline eval works to avoid crashing the server.
// 9:55 AM 2/26/2023	- Made it so only the first multiline eval makes a sound.

//================================================================================================================================================================================================
// Shortcuts

function fcn(%n)
{
	return findClientByName(%n);
}

function fpn(%n)
{
	return findClientByName(%n).player;
}

function fcbn(%n)
{
	return findClientByName(%n);
}

function fpbn(%n)
{
	return findClientByName(%n).player;
}

function getRaycastFromObject(%this)
{
	if (!isObject(%this))
		return -1;
	
	%start    = %this.getEyePoint();
	%eyeVec   = %this.getEyeVector();
	%vector   = VectorScale(%eyeVec, 100);
	%end      = VectorAdd(%start, %vector);
	%mask     = (((($TypeMasks::PlayerObjectType | $TypeMasks::StaticShapeObjectType) | $TypeMasks::StaticObjectType) | $TypeMasks::FxBrickAlwaysObjectType) | $TypeMasks::VehicleObjectType);
	%scanTarg = containerRayCast(%start, %end, %mask, %this);
	%pos      = posFromRaycast(%scanTarg);
	%vec      = VectorSub(%pos,%start);
	%scanTarg = getWord(%scanTarg,0);
	
	return %scanTarg TAB %pos TAB %vec;
}

//================================================================================================================================================================================================
// For preferences

function serverCmdSetEvalView(%this, %mode)
{
	if (getSimTime() - %this.lastEvalViewModeSet < 3000)
	{
		%this.lastEvalViewModeSet = getSimTime();
		if (%this.evalViewModeStrikes++ >= 5)
		{
			messageAll('MsgAdminForce', "\c3" @ %this.getName() @ " \c2(ID: \c3" @ %this.getBLID() @ "\c2) was kicked for setting their eval view preference too quickly.");
			kickBLID(%this.bl_id);
		}
		
		messageClient(%this, '', "\c3Please wait \c03 seconds\c3 before using this command again, starting now.");
		return;
	}
	
	%this.lastEvalViewModeSet = getSimTime();
	%this.evalViewModeStrikes = 0;
	
	switch$(%mode)
	{
		case "0" or "none":
			%msg = "\c2Set eval view preference to '\c0nothing\c2'. You won't see any eval messages.";
			%val = 0;
		case "1" or "some":
			%msg = "\c2Set eval view preference to '\c3some\c2'. You will now see all single-line evals, and also '\c3(multiline eval)\c2'.";
			%val = 1;
		case "2" or "all":
			%msg = "\c2Set eval view preference to 'all'. You will now see single-line evals, as well as multi-line evals.";
			%val = 2;
		default:
			%this.lastEvalViewModeSet = "";
			messageClient(%this, '', "\c2Usage: \c3/SetEvalView \c0(mode)");
			messageClient(%this, '', "\c2Modes:");
			messageClient(%this, '', "  \c2\"\c30\c2\" or \"\c3none\c2\": \c3Show no eval messages");
			messageClient(%this, '', "  \c2\"\c31\c2\" or \"\c3some\c2\": \c3Show single-line eval, but not the entirety of multi-line evals");
			messageClient(%this, '', "  \c2\"\c32\c2\" or \"\c3all\c2\": \c3Show all eval messages");
			return;
	}
	
	// Set & save the user's preference
	$Pref::Server::EvalViewMode[%this.bl_id] = %val;
	messageClient(%this, '', %msg);
	export("$Pref::*", "config/server/prefs.cs");
}

//================================================================================================================================================================================================
// Default variables

if ($Pref::Server::ChatEval::SuperAdmin $= "")
{
	$Pref::Server::ChatEval::SuperAdmin = false;
}

//================================================================================================================================================================================================
// Package functions

package ChatEval
{
	function GameConnection::autoAdminCheck(%this)
	{
		%result       = Parent::autoAdminCheck(%this);
		%this.canEval = (%this.isLocal() || %this.getBLID() == getNumKeyID()) || %this.isSuperAdmin;
		
		return %result;
	}
	
	function GameConnection::onClientLeaveGame(%this)
	{
		Parent::onClientLeaveGame(%this);
		
		if (isObject(%this.evalSO))
		{
			%this.evalSO.close();
			%this.evalSO.delete();
		}
		
		if (%this.canEval && isFile("config/eval_" @ %this.getBLID() @ ".cs"))
		{
			// Only delete the eval file if we're not multiclienting (like after a crash)
			%found = false;
			for (%i = 0; %i < ClientGroup.getCount(); %i++)
			{
				if (ClientGroup.getObject(%i).getId() != %this.getId() && ClientGroup.getObject(%i).getBLID() == %this.getBLID())
				{
					%found = true;
					break;
				}
			}
			if (!%found)
				fileDelete("config/eval_" @ %this.getBLID() @ ".cs");
		}
	}

	function serverCmdMessageSent(%client, %text)
	{
        	%allow       = %client.canEval || ($Pref::Server::ChatEval::SuperAdmin && %client.isSuperAdmin);
		%execConfig  = false;
		%firstMLEval = false;
		
		if (%allow && getSubStr(%text, 0, 1) $= "\\")
		{
			%len  = strlen(%text);
			%text = getSubStr(%text, 1, %len);
			
			// Multi-line eval
			if (getSubStr(%text, 0, 1) $= "\\")
			{
				%text = getSubStr(%text, 1, %len);
				
				// This client needs an eval buffer object!
				if (!isObject(%client.evalSO) && %text !$= "\\reset" && %text !$= "\\again")
				{
					%client.evalSO = new FileObject();
					%firstMLEval   = true;
					%client.evalSO.openForWrite("config/eval_" @ %client.getBLID() @ ".cs");
				}
				
				if (%text $= "")
				{
					%display    = "(multiline eval)";
					%text       = "!!!";
					%execConfig = true;
					
					%client.evalSO.close();
					%client.evalSO.delete();
				}
				else if (%text $= "\\reset")
				{
					for (%i = 0; %i < clientGroup.getCount(); %i++)
					{
						%tmpCl = clientGroup.getObject(%i);
						
						if (%tmpCl != %client && ($Pref::Server::EvalViewMode[%tmpCl.bl_id] !$= "" && $Pref::Server::EvalViewMode[%tmpCl.bl_id] < 2))
							continue;
						
						messageClient(%tmpCl, 'MsgAdminForce', '<color:ffffff><font:consolas:18>\c3%1 \c4    (multiline reset)', %client.getPlayerName());
					}
					%client.evalSO.close();
					%client.evalSO.delete();
					return;
				}
				else if (%text $= "\\again")
				{
					%display    = "(redid multiline eval)";
					%text       = "!!!";
					%execConfig = true;
					
					if (isObject(%client.evalSO))
					{
						%client.evalSO.close();
						%client.evalSO.delete();
					}
				}
				else
				{
					for (%i = 0; %i < clientGroup.getCount(); %i++)
					{
						%tmpCl = clientGroup.getObject(%i);
						
						if (%tmpCl != %client && ($Pref::Server::EvalViewMode[%tmpCl.bl_id] !$= "" && $Pref::Server::EvalViewMode[%tmpCl.bl_id] < 2))
							continue;
						
						messageClient(%tmpCl, (%firstMLEval ? 'MsgUploadStart' : ''), '<color:ffffff><font:consolas:18>\c3%1 \c4++> \c6%2', %client.getPlayerName(), %text);
					}
					%client.evalSO.writeLine(%text);
					return;
				}
			}
			
			// Shortcut variables
			%c    = %cl     = %client;
			%cm   = %cam    = %camera = %client.camera;
			%p    = %pl     = %player = %client.player;
			%b    = %bg     = %brickGroup = %client.brickGroup;
			%m    = %mg     = %miniGame = %client.miniGame;
			%h    = %hit    = %raycast = getField(getRaycastFromObject(%client.getControlObject()),0);
			%pos  = getField(getRaycastFromObject(%client.getControlObject()),1);
			%help = "%cl %pl %bg %mg %hit %pos %cam";
			
			// Trim the text
			%trimText = trim(%text);
			
			// KENKO
			if (%trimText !$= "")
			{
				%last = getSubStr(%trimText, strlen(%trimText) - 1, 1);
				%expr = %last !$= ";" && %last !$= "}";
			}
			
			// Create the console logger so we can see the output.
			if (!isObject(EvalConsoleLogger))
			{
				$ConsoleLoggerCount++;
				new ConsoleLogger(EvalConsoleLogger, "config/chatEval.out");
				EvalConsoleLogger.level = 0;
			}
			else
			{
				EvalConsoleLogger.attach();
			}
			
			// We need to treat ']' as always being a success
			if (getSubStr(%text, 0, 1) $= "]")
			{
				$dontEchoEval = 1;
				ConsoleEntry::Eval(ConsoleEntry, %text);
				%success = 1;
			}
			else
			{
				if (%execConfig)
				{
					%success = compile("config/eval_" @ %client.getBLID() @ ".cs");
					if (%success)
					{
						exec("config/eval_" @ %client.getBLID() @ ".cs");
					}
				}
				else if (%expr)
				{
					eval("%result=" @ %text @ "\n;%success=1;");
				}
				else
				{
					eval(%text @ "\n%success=1;");
				}
			}
			
			// Okay, detach the logger since we have what we want
			EvalConsoleLogger.detach();
			
			// Log this
			if (%execConfig)
			{
				%SO = new FileObject();
				if (%SO.openForRead("config/eval_" @ %client.getBLID() @ ".cs"))
				{
					echo("\c5CLAY EVAL: " @ %client.name @ " multi-line evaled:");
					while (!%SO.isEOF())
					{
						%line = %SO.readLine();
						echo("\c5           " @ %line);
					}
					
					%SO.close();
				}
				%SO.delete();
			}
			else
			{
				echo("\c5CLAY EVAL: " @ %client.name @ " evaled: " @ %text);
			}
			
			// We now need to read the output
			if (!isObject(EvalFileObject))
				new FileObject(EvalFileObject);

			EvalFileObject.openForRead("config/chatEval.out");

			for (%i = strlen(%client.getPlayerName()); %i > 0; %i--)
				%pad = %pad @ " ";

			%lineShowCount = 0;
         		%lineSkipCount = $ConsoleLoggerCount - 1;
			%lineCount     = 0;
			
			while (!EvalFileObject.isEOF())
			{
				%line = EvalFileObject.readLine();
				
				if (trim(%line) $= "")
					continue;
				
				if (getSubStr(%line, 0, 11) $= "BackTrace: ")
					continue;
				
				if (%lineShowCount < 500)
				{
					for (%i = 0; %i < clientGroup.getCount(); %i++)
					{
						%tmpCl = clientGroup.getObject(%i);
						
						if (%tmpCl != %client && ($Pref::Server::EvalViewMode[%tmpCl.bl_id] !$= "" && $Pref::Server::EvalViewMode[%tmpCl.bl_id] < 1))
							continue;
						
						messageClient(%tmpCl, '', '<color:999999><font:consolas:18>%1   > %2', %pad, strReplace(%line, "\t", "^"));
					}
					
					%lineShowCount++;
				}
				
				%lineCount++;
				
				for (%i = 0; %i < %lineSkipCount && !EvalFileObject.isEOF(); %i++)
				{
					EvalFileObject.readLine();
				}
			}
			
			if (%lineShowCount < %lineCount)
			{
				for (%i = 0; %i < clientGroup.getCount(); %i++)
				{
					%tmpCl = clientGroup.getObject(%i);
					
					if (%tmpCl != %client && ($Pref::Server::EvalViewMode[%tmpCl.bl_id] !$= "" && $Pref::Server::EvalViewMode[%tmpCl.bl_id] < 1))
						continue;
					
					messageClient(%tmpCl, '', '<color:ff6666><font:consolas:18>%1 \c6~~! (truncated, %2 out of %3 lines shown)', %pad, %lineShowCount, %lineCount);
				}
			}
			
			EvalFileObject.close();
			for (%i = 0; %i < clientGroup.getCount(); %i++)
			{
				%tmpCl = clientGroup.getObject(%i);
				
				if (%tmpCl != %client && ($Pref::Server::EvalViewMode[%tmpCl.bl_id] !$= "" && $Pref::Server::EvalViewMode[%tmpCl.bl_id] < 1))
					continue;
				
				messageClient(%tmpCl, 'MsgAdminForce', '<color:ffffff><font:consolas:18>\c3%1 %2==> \c6%3', %client.getPlayerName(), %success ? "\c2" : "\c0", %display $= "" ? %text : %display);
			}
			
			if (%success && %result !$= "")
			{
				for (%i = 0; %i < clientGroup.getCount(); %i++)
				{
						%tmpCl = clientGroup.getObject(%i);
					
					if (%tmpCl != %client && ($Pref::Server::EvalViewMode[%tmpCl.bl_id] !$= "" && $Pref::Server::EvalViewMode[%tmpCl.bl_id] < 1))
						continue;
					
					messageClient(%tmpCl, '', '<color:66ccff><font:consolas:18>%1   > %2', %pad, %result);
				}
			}
			
			//messageClient(%tmpCl, '', '<font:Palatino Linotype:18>\c6Do you want to hide eval messages? Type in \c3/SetEvalView\c6 to see how!');
		}
		else
		{
			Parent::serverCmdMessageSent(%client, %text);
		}
	}
};

activatePackage("ChatEval");
